'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var effector = require('effector');

const QUEUE = 'QUEUE';
const RACE = 'RACE';
const TAKE_EVERY = 'TAKE_EVERY';
const TAKE_FIRST = 'TAKE_FIRST';
const TAKE_LAST = 'TAKE_LAST';

class ReEffectError extends Error {
  constructor(message) {
    super(message);
    delete this.stack;
  }

}
class CancelledError extends ReEffectError {
  constructor(strategy = 'cancel') {
    super('Cancelled due to "' + strategy + '"' + ({
      [TAKE_FIRST]: `, there are already running effects`,
      [TAKE_LAST]: `, new effect was added`,
      [RACE]: `, other effect won race`,
      cancel: `, cancel all already running effects`
    }[strategy] || `, but should not happen...`));
  }

}
class LimitExceededError extends ReEffectError {
  constructor(limit, running) {
    super('Cancelled due to limit of "' + limit + '" exceeded,' + 'there are already ' + running + ' running effects');
  }

}
class TimeoutError extends ReEffectError {
  constructor(timeout) {
    super('Cancelled due to timeout of "' + timeout + '"ms');
  }

}

const assign = Object.assign;
const own = ({
  graphite: owner
}, links) => {
  for (const {
    graphite
  } of links) {
    graphite.family.type = 'crosslink';
    graphite.family.owners.push(owner);
    owner.family.links.push(graphite);
  }
};
const setMeta = (unit, field, value) => {
  unit.graphite.meta[field] = value;
};
const getForkPage = val => val.scope;
const read = scope => $store => scope ? scope.getState($store) : $store.getState();

const cancellable = (promise, abort, timeout) => {
  let cancel;
  const cancelable = new Promise((_, reject) => {
    let timeoutId;

    const rejectWith = Cls => arg => {
      clearTimeout(timeoutId);
      reject(new Cls(arg));
      abort && abort();
    };

    cancel = rejectWith(CancelledError);

    if (typeof timeout === 'number') {
      timeoutId = setTimeout(rejectWith(TimeoutError), timeout, timeout);
    }
  });
  return assign(Promise.race([promise, cancelable]), {
    cancel
  });
};
const defer = () => {
  const deferred = {};
  deferred.req = new Promise((resolve, reject) => {
    deferred.rs = resolve;
    deferred.rj = reject;
  });
  deferred.req.catch(() => {});
  return deferred;
};
const createRunning = () => {
  const cancelAllEv = effector.createEvent({
    sid: 'internal/cancelAll'
  });
  const $running = effector.createStore([], {
    sid: 'internal/$running',
    serialize: 'ignore'
  });
  $running.watch(cancelAllEv, (runs, strategy) => runs.forEach(p => p.cancel(strategy)));

  const push = (promise, scope) => {
    read(scope)($running).push(promise);
  };

  const unpush = (promise, scope) => {
    if (!promise) return;
    read(scope)($running).splice(read(scope)($running).indexOf(promise), 1);
  };

  const cancelAll = (strategy, scope) => {
    read(scope)($running).forEach(p => p.cancel(strategy));
  };

  return {
    $running,
    push,
    unpush,
    cancelAll,
    cancelAllEv
  };
};

const patchInstance = (instance, {
  cancelled,
  cancel,
  feedback,
  inFlight,
  pending
}) => {
  assign(instance, {
    cancelled,
    cancel,
    inFlight,
    pending
  });
  own(instance, [cancelled, cancel]);
  setMeta(cancelled, 'needFxCounter', 'dec');

  instance.create = (paramsOrConfig, [strategyOrConfig]) => {
    const config = paramsOrConfig && (paramsOrConfig.strategy || paramsOrConfig.timeout) ? paramsOrConfig : strategyOrConfig && (strategyOrConfig.strategy || strategyOrConfig.timeout) ? strategyOrConfig : {
      strategy: strategyOrConfig
    };
    const req = defer();
    effector.launch(instance, {
      params: config === paramsOrConfig ? config.params : paramsOrConfig,
      args: {
        strategy: config.strategy,
        timeout: config.timeout
      },
      req
    });
    return req.req;
  };

  if (feedback) {
    const feedbackStep = effector.step.compute({
      fn(data, _, stack) {
        data.strategy = stack.parent.parent.value.strategy;
        return data;
      }

    });
    instance.done.graphite.seq.push(feedbackStep);
    instance.fail.graphite.seq.push(feedbackStep);
  }
};

const patchRunner = (runner, runnerScope) => {
  const runs = createRunning();
  assign(runner.scope, runnerScope);
  runner.seq = seq(runnerScope.anyway, runs);
  runnerScope.cancel.watch(() => runs.cancelAllEv());
};

const seq = (anyway, runs) => [effector.step.compute({
  safe: true,
  filter: false,
  priority: 'effect',

  fn(upd, scopeArg, stack) {
    const scope = scopeArg;
    let handler = scope.handler;

    if (getForkPage(stack)) {
      const handlerArg = getForkPage(stack).handlers[scope.handlerId];
      if (handlerArg) handler = handlerArg;
    }

    upd.handler = handler;
    return upd;
  }

}), effector.step.run({
  fn({
    params,
    args,
    req,
    handler
  }, runScope, {
    scope
  }) {
    const {
      feedback,
      limit,
      cancelled,
      inFlight
    } = runScope;
    let {
      strategy,
      timeout
    } = runScope;
    strategy = args && args.strategy || strategy;
    timeout = args && args.timeout || timeout;
    const runnerScope = {
      params,
      strategy,
      timeout,
      feedback,
      push: runs.push,
      unpush: runs.unpush,
      cancelAll: runs.cancelAll,
      inFlight,
      scope,
      $running: runs.$running
    };
    const go = run(runnerScope, handler, fin(runnerScope, anyway, 0, req.rs), fin(runnerScope, anyway, 1, req.rj), fin(runnerScope, cancelled, 2, req.rj));
    const running = read(scope)(runs.$running);

    if (running.length >= limit) {
      return go(new LimitExceededError(limit, running.length));
    }

    if (strategy === TAKE_FIRST && running.length > 0) {
      return go(new CancelledError(strategy));
    }

    if (strategy === TAKE_LAST && running.length > 0) {
      runs.cancelAll(strategy, scope);
    }

    if (strategy === QUEUE && running.length > 0) {
      const promise = cancellable(Promise.all(running.map(p => p.catch(() => {}))));
      runs.push(promise, scope);
      return promise.then(() => (runs.unpush(promise, scope), go()), error => (runs.unpush(promise, scope), go(error)));
    }

    go();
  }

})];

const run = ({
  params,
  push,
  timeout,
  scope
}, handler, onResolve, onReject, onCancel) => immediatelyCancelError => {
  if (immediatelyCancelError) {
    return onCancel()(immediatelyCancelError);
  }

  let cancel;
  let result;

  try {
    result = handler(params, abort => cancel = abort);
  } catch (error) {
    return onReject()(error);
  }

  if (Object(result) === result && typeof result.then === 'function') {
    const promise = cancellable(result, cancel, timeout);
    push(promise, scope);
    return promise.then(onResolve(promise), error => error instanceof CancelledError ? onCancel(promise)(error) : onReject(promise)(error));
  }

  return onResolve()(result);
};

const internalFinally = effector.createEvent({
  named: 'internalFinally'
});
setMeta(internalFinally, 'needFxCounter', 'dec');

const fin = ({
  params,
  unpush,
  strategy,
  feedback,
  cancelAll,
  inFlight,
  scope,
  $running
}, target, type, fn) => promise => data => {
  unpush(promise, scope);
  const runningCount = read(scope)($running).length;
  const targets = [sidechain];
  const payloads = [[fn, data]];

  if (type === 2 || !runningCount || strategy === RACE) {
    const body = type === 0 ? {
      params,
      result: data,
      status: 'done'
    } : type === 1 ? {
      params,
      error: data,
      status: 'fail'
    } : {
      params,
      error: data
    };
    if (feedback) body.strategy = strategy;
    targets.unshift(target);
    payloads.unshift(body);

    if (strategy === RACE && type !== 2) {
      cancelAll(strategy, scope);
    }
  } else if (runningCount && (strategy === TAKE_EVERY || strategy === QUEUE)) {
    targets.push(internalFinally);
  }

  if (runningCount && (type !== 2 || strategy !== RACE)) {
    effector.launch({
      scope: scope,
      target: [inFlight],
      params: [runningCount]
    });
  } else {
    targets.unshift(inFlight);
    payloads.unshift(runningCount);
  }

  effector.launch({
    target: targets,
    params: payloads,
    defer: true,
    scope
  });
};

const sidechain = effector.createNode({
  node: [effector.step.run({
    fn([fn, value]) {
      fn(value);
    }

  })],
  meta: {
    op: 'fx',
    fx: 'sidechain'
  }
});

const createReEffectFactory = (createEffect = effector.createEffect) => (nameOrConfig, maybeConfig) => {
  const instance = createEffect(nameOrConfig, maybeConfig);
  const cancelled = effector.createEvent({
    named: 'cancelled'
  });
  const cancel = effector.createEvent({
    named: 'cancel'
  });
  const inFlightInternal = effector.createStore(0, {
    named: 'reeffectInFlight'
  }).on(instance, s => s + 1);
  const pendingInternal = inFlightInternal.map({
    fn: amount => amount > 0,
    named: 'reeffectPending'
  });
  const config = maybeConfig ? maybeConfig : nameOrConfig && typeof nameOrConfig === 'object' ? nameOrConfig : {};
  const scope = {
    strategy: config.strategy || TAKE_EVERY,
    feedback: config.feedback || false,
    limit: config.limit || Infinity,
    timeout: config.timeout,
    cancelled,
    cancel,
    inFlight: inFlightInternal,
    pending: pendingInternal,
    anyway: instance.finally
  };
  patchRunner(instance.graphite.scope.runner, scope);
  patchInstance(instance, scope);
  return instance;
};

const createReEffect = createReEffectFactory();

exports.CancelledError = CancelledError;
exports.LimitExceededError = LimitExceededError;
exports.QUEUE = QUEUE;
exports.RACE = RACE;
exports.ReEffectError = ReEffectError;
exports.TAKE_EVERY = TAKE_EVERY;
exports.TAKE_FIRST = TAKE_FIRST;
exports.TAKE_LAST = TAKE_LAST;
exports.TimeoutError = TimeoutError;
exports.createReEffect = createReEffect;
exports.createReEffectFactory = createReEffectFactory;
//# sourceMappingURL=index.js.map
